package edu.stanford.hci.dtools.serial;

import gnu.io.CommPortIdentifier;
import gnu.io.NoSuchPortException;
import gnu.io.PortInUseException;
import gnu.io.SerialPort;
import gnu.io.SerialPortEvent;
import gnu.io.SerialPortEventListener;
import gnu.io.UnsupportedCommOperationException;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.Timer;
import java.util.TimerTask;
import java.util.TooManyListenersException;

import com.illposed.osc.OSCListener;
import com.illposed.osc.OSCMarshalIn;
import com.illposed.osc.OSCPacket;
import com.illposed.osc.utility.OSCByteArrayToJavaConverter;
import com.illposed.osc.utility.OSCPacketDispatcher;

public abstract class DtoolsSerialPort implements Runnable, OSCMarshalIn, SerialPortEventListener {

	public static final int DEFAULT_BAUD_RATE = 115200;

	public abstract void serialEvent(SerialPortEvent event);

	public static final int DEFAULT_DATA_BITS = SerialPort.DATABITS_8;
	public static final int DEFAULT_STOP_BITS = SerialPort.STOPBITS_1;
	public static final int DEFAULT_PARITY = SerialPort.PARITY_NONE;
	/**
	 * We occasionally loose bytes going out to the d.tools hardware boards.
	 * When this flag is set to true, outgoing messages are padded with two null bytes
	 * so the embedded OSC message parser can detect errors in time request a resend
	 * WARNING: it's not guaranteed that other OSC parsers will gracefully handle extra null bytes.
	 */
	public static final boolean PAD_MESSAGES = true;
	static CommPortIdentifier portId;
	static Enumeration portList;
	protected InputStream inputStream;
	protected OutputStream outputStream;
	protected SerialPort serialPort;
	protected Thread readThread;
	
	protected static String portName;
	protected static int portBaudRate = DEFAULT_BAUD_RATE;
	protected static int portDataBits = DEFAULT_DATA_BITS;
	protected static int portStopBits = DEFAULT_STOP_BITS;
	protected static int portParity = DEFAULT_PARITY;
	protected OSCByteArrayToJavaConverter converter = new OSCByteArrayToJavaConverter();
	protected OSCPacketDispatcher dispatcher = new OSCPacketDispatcher();
	private Timer timer;
	
	public DtoolsSerialPort() {
		super();
	}

	/**
	 * 
	 */
	protected boolean openCommPort(String pName, int baudRate, int dataBits, int stopBits, int parity) {
	    portId=null;
	   	serialPort=null;
	   	try {
	   		portName = pName;
	   		portBaudRate = baudRate;
	   		portDataBits = dataBits;
	   		portStopBits = stopBits;
	   		portParity = parity;
	   		portId = CommPortIdentifier.getPortIdentifier(portName);
	   		System.out.println("Found port: "+portName);
	   	} catch(NoSuchPortException e) {
	   		System.out.println("port " + portName + " not found.");
	   		return false;
	   	}
	    initParser();	
	   	
		   try {
			    serialPort = (SerialPort) portId.open("OSCSerialPort", 2000);
			} catch (PortInUseException e) {
				System.err.println("PortInUseException!\n");
				return false;
			}
			try {
			    inputStream = serialPort.getInputStream();
			} catch (IOException e) {
				System.err.println("IOException in serialPort.getInputStream()!\n");
				return false;
			}
			try {
				outputStream = serialPort.getOutputStream();
			} catch (IOException e) {
				System.err.println("IOException in serialPort.getOutputStream()!\n");
				return false;
			}
			try {
			    serialPort.addEventListener(this);
			    serialPort.notifyOnDataAvailable(true);
				serialPort.notifyOnOutputEmpty(true);
			} catch (TooManyListenersException e) {
				System.err.println("TooManyListenersException!\n");
				return false;
			}
			
			try {
			    serialPort.setSerialPortParams(baudRate, dataBits, 
							   stopBits, 
							   parity);
			} catch (UnsupportedCommOperationException e) {
				System.err.println("UnsupportedCommOperationException!\n");
				return false;
			}
			return true;
	   }

	abstract void initParser();

	public void run() {
		timer = new Timer();
		//timer.scheduleAtFixedRate(new CheckSerialPortTimerTask(),10000,10000);
		try {
		    Thread.sleep(20000);
		} catch (InterruptedException e) {
			System.err.println("InterruptedException in OSCSerialPort.run()");
		}
		
	}

	
	/**
	 * Send an OSC packet out our open serial port
	 * @param aPacket The OSC packet that should be sent
	 * @throws IOException
	 */
	public void send(OSCPacket aPacket) throws IOException {
	 //if we have no port, try opening again
	 try {
		 CommPortIdentifier.getPortIdentifier(portName);
	 } catch(NoSuchPortException e) {
		 System.err.println("COM port does not exist  - not sending OSC message.");
			 throw new IOException();
	 }
	
	int checksum=0;
		// get byte array of OSC message
		byte[] byteArray = aPacket.getByteArray();
		// package into serial format:
		int length = byteArray.length;
		if(length<=0) {
			//we have an error in the byte array
			System.err.println("OSCSerialPort.send: bad byte array");
			return;
		}
		
		byte[] serialByteArray;
		if(PAD_MESSAGES) {
			serialByteArray = new byte[byteArray.length+5];
		} else {
			serialByteArray = new byte[byteArray.length+3];
		}
		
		// add header
		serialByteArray[0]=(byte)0xBE;
		// add length
		serialByteArray[1]=(byte)byteArray.length;
		// copy message
		for(int i=0; i<byteArray.length; i++) {
			serialByteArray[2+i]=byteArray[i];
			checksum=((checksum + byteArray[i])% 256);
		}
		// add checksum
		serialByteArray[byteArray.length+2]=(byte)checksum;
		//some zero bytes to pad
		if(PAD_MESSAGES) {
			serialByteArray[byteArray.length+3]=0;
			serialByteArray[byteArray.length+4]=0;
		}
		// send message
		//System.out.println("OSCSerialPort: Sending "+serialByteArray.length+" bytes out.");
		try {
			outputStream.write(serialByteArray);
		} catch (Exception e) {
			System.err.println("Encountered exception while trying to send serial message.");
			e.printStackTrace();
			if (serialPort!=null) serialPort.close();
			// try to reopen
			if(!openCommPort(portName,portBaudRate,portDataBits,portStopBits,portParity)) {
			 System.err.println("Unsuccesssfully tried to reopen COM port - not sending OSC message.");
			 if(serialPort!=null) serialPort.close();
			 throw new IOException();
			} else  {
				try {
					outputStream.write(serialByteArray);
				} catch (Exception f ) {
					System.err.println("Failed a second time.");
					throw new IOException();
				}
			}
		 
	
		}
	}

	/**
	 * Register the listener for incoming OSCPackets addressed to an Address
	 * @param anAddress  the address to listen for
	 * @param listener   the object to invoke when a message comes in
	 */
	public void addListener(String anAddress, OSCListener listener) {
		dispatcher.addListener(anAddress, listener);
	}

	/**
	 * Un-register all existing listeners
	 * @author bjoern
	 * july 2005
	 */
	public void removeAllListeners() {
		dispatcher.removeAllListeners();
	}

	public boolean removeListener(String anAddress) {
		return dispatcher.removeListener(anAddress);
	}

	public boolean changeListenerAddress(String oldAddress, String newAddress) {
		return dispatcher.changeListenerAddress(oldAddress, newAddress);
	}

	public OSCListener getListener(String anAddress) {
		return dispatcher.getListener(anAddress);
	}

	/**
	 * Close the socket and free-up resources. It's recommended that clients call
	 * this when they are done with the port.
	 */
	public void close() {
		if (serialPort!=null) serialPort.close();
	}

	/**
	 * @see java.lang.Object#finalize()
	 */
	protected void finalize() throws Throwable {
		if(serialPort!=null) serialPort.close();
		this.finalize();
	}
	
	class CheckSerialPortTimerTask extends TimerTask {
		@Override
		public void run() {
			//set this timer to be expired
			System.err.print("Checking Serial port...");
			try{
				byte[] byteArray = new byte[1];
				byteArray[0]=0x20; //send some dummy byte
				outputStream.write(byteArray);
				System.err.println("ok.");
			}
			catch(IOException e) {
				System.err.print("error...");
				if(serialPort!=null) {
					serialPort.close();
				}
				if(openCommPort(portName,portBaudRate,portDataBits,portStopBits,portParity))
					System.err.println("recovered.");
				else
					System.err.println("could not recover.");
			}
		}
	}

}