/**
 * 
 */
package edu.stanford.hci.dtools.serial;

import java.io.IOException;

import com.illposed.osc.OSCMarshalIn;
import com.illposed.osc.OSCPacket;

import gnu.io.SerialPortEvent;



/**
 * Serial input/output of OSC messages.
 * Uses Pascal Stang's OSC serial format:
 * [0xBE = header byte][size byte][...message...][8bit checksum]
 * <p>
 * This software is distributed under the <a href="http://hci.stanford.edu/research/copyright.txt">
 * BSD License</a>.
 * </p>
 * 
 * @author <a href="http://bjoern.org/">Bjoern Hartmann</a> ( bjoern(AT)stanford.edu )
 */
public class OSCSerialPort extends DtoolsSerialPort implements OSCMarshalIn {
	
	private int oscRxNextOp;
	private int oscRxMsgSize;
	private int oscRxReadBytes;
	private byte[] oscRxData;
	
	static final int OSC_RXOP_WAITFORSTART = 0;
	static final int OSC_RXOP_READSIZE = 1;
	static final int OSC_RXOP_READDATA = 2;
	static final int OSC_RXOP_READCHKSUM = 3;

	
	// Constructors
 	public OSCSerialPort(String portName) {
 		super();
 		if(openCommPort(portName,DEFAULT_BAUD_RATE,DEFAULT_DATA_BITS,DEFAULT_STOP_BITS,DEFAULT_PARITY)) {
 			readThread = new Thread(this);
 			readThread.start();
 		}
 	}
 	public OSCSerialPort(String portName, int baudRate) {
 		super();
 		if(openCommPort(portName,baudRate,DEFAULT_DATA_BITS,DEFAULT_STOP_BITS,DEFAULT_PARITY)) {
 			readThread = new Thread(this);
 			readThread.start();
 		}
 	}
 	public OSCSerialPort(String portName, int baudRate, int dataBits, int stopBits, int parity) {
 		super();
 		if(openCommPort(portName,baudRate,dataBits,stopBits,parity)) {
 			readThread = new Thread(this);
 			readThread.start();
 		}
 	}

 	
    /**
	 * Handles incoming serial data.
	 * Passes bytes one by one to buildOSCByteBuffer().
	 * Ignores all other serial events.
	 */
	@Override
	public void serialEvent(SerialPortEvent event) {
		//System.out.println("SerialEvent came in.");
    	switch (event.getEventType()) {
    	case SerialPortEvent.BI:
    	case SerialPortEvent.OE:
    	case SerialPortEvent.FE:
    	case SerialPortEvent.PE:
    	case SerialPortEvent.CD:
    	case SerialPortEvent.CTS:
    	case SerialPortEvent.DSR:
    	case SerialPortEvent.RI:
    	case SerialPortEvent.OUTPUT_BUFFER_EMPTY:
    		break;
    	case SerialPortEvent.DATA_AVAILABLE:
    		try {
    			while (inputStream.available() > 0) {
    				buildOSCByteBuffer(inputStream.read());
    			}
    		} catch (IOException e) {
    			System.err.println("IOException in serialEvent!\n");
    		}
    		break;
    	}
		
	}
	/**
	 * Piece together an OSCByteBuffer one byte at a time.
	 * Expects OSC header 0xBE, size byte, and 8bit checksum.
	 * 
	 * @param c = current byte read from serial stream
	 */
	protected void buildOSCByteBuffer(int c) {
		switch(oscRxNextOp)
		{
		case OSC_RXOP_WAITFORSTART:
			if(c==0xBE) {
				oscRxNextOp = OSC_RXOP_READSIZE;
			}
			break;
		case OSC_RXOP_READSIZE:
			oscRxMsgSize = c;
			oscRxReadBytes = 0;
			oscRxData = new byte[oscRxMsgSize+1];
			oscRxNextOp = OSC_RXOP_READDATA;
			break;
		case OSC_RXOP_READDATA:
			try {
			oscRxData[oscRxReadBytes++] = (byte)c;
			if(oscRxReadBytes == oscRxMsgSize) {
			oscRxData[oscRxReadBytes] = 0; //terminate string
				oscRxNextOp = OSC_RXOP_READCHKSUM;
			}
			} catch (java.lang.ArrayIndexOutOfBoundsException e) {
				//something went wrong (i.e., got an erroneous 0xBE- throw the message away
				oscRxNextOp = OSC_RXOP_WAITFORSTART;
			}
			break;
		case OSC_RXOP_READCHKSUM:
			short oscRxChecksum = 0;
			for (int i=0; i<oscRxMsgSize; i++) {
				oscRxChecksum=(short)((oscRxChecksum + (short)(0x000000FF &(int)oscRxData[i]))% 256);
			}
			if(oscRxChecksum == c) {
				// checksum matched and we're done with this msg
				// -> fire off registered listeners
				
				OSCPacket oscPacket = converter.convert(oscRxData,oscRxMsgSize);
				//System.out.println("received valid OSC message.");
				dispatcher.dispatchPacket(oscPacket);
				oscRxNextOp = OSC_RXOP_WAITFORSTART;
			} else {
				// mismatch - throw this message away
				System.err.println("Rcv corrupted OSC msg: calculated checksum "+(int)oscRxChecksum+" should have been: "+(int)c+". Skipping.");
				oscRxNextOp = OSC_RXOP_WAITFORSTART;
			}
			break;
		}
	}
	@Override
	void initParser() {
		oscRxNextOp = OSC_RXOP_WAITFORSTART;
		
	}

	
}
