package edu.stanford.hci.exemplar;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Timer;

import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.ui.plugin.AbstractUIPlugin;
import org.osgi.framework.BundleContext;

import com.illposed.osc.OSCMarshalIn;
import com.illposed.osc.OSCPortIn;
import com.thoughtworks.xstream.XStream;

import edu.stanford.hci.dtools.io.make.MakeControllerOSCPort;
import edu.stanford.hci.dtools.serial.OSCSerialPort;
import edu.stanford.hci.dtools.serial.WiTiltSerialPort;
import edu.stanford.hci.exemplar.flashserver.FlashServer;
import edu.stanford.hci.exemplar.output.FlashOutput;
import edu.stanford.hci.exemplar.output.OscOutput;
import edu.stanford.hci.exemplar.preferences.PreferenceConstants;

/**
 * The activator class controls the plug-in life cycle
 *
 * Exemplar
 * Copyright (c) 2004-2007, Stanford University
 * All rights reserved.
 *
 * <p>
 * This software is distributed under the <a href="http://hci.stanford.edu/research/copyright.txt">
 * BSD License</a>.
 * </p>
 * 
 * 
 * @author <a href="http://bjoern.org/">Bjoern Hartmann</a> ( bjoern(AT)stanford.edu )
 * @author Leith Abdulla
 * @author Manas Mittal
 * 
 */
public class Activator extends AbstractUIPlugin implements PropertyChangeListener {

	// The plug-in ID
	public static final String PLUGIN_ID = "edu.stanford.hci.exemplar";
	

	// The shared instance
	private static Activator plugin;
	private Timer timer;
	private final UpdateDataTask updateData = new UpdateDataTask();
	private final RedrawDataTask redrawData = new RedrawDataTask();
	
	
	//this could be a DtoolsSerialPort, a WiTiltSerialPort, or a OSCPortIn
	private OSCMarshalIn inputPort=null;
	
	private  Map<String,Sensor> sensors = new HashMap<String,Sensor>(); //Sensor is our data model
	private Map<String,EventSet> eventSets = new HashMap<String,EventSet>();
	private Map<List<String>,EventSet> eventMultiSets = new HashMap<List<String>,EventSet>();
	public final PropertyChangeSupport pcs;
	
	public static EventSet ActiveEventSet = null;
	private final List<String> activeSensors = new ArrayList<String>();

	private long slowPeriod = 1000/15; //15Hz - screen redraw frequency
	private long fastPeriod = 1000/50; //50Hz - sensor poll frequency
	long delay = 1000; //wait a second before starting timers
	//private int currentTime =0;

	private String singleActiveSensor = "/adc0";
	private EventRegion clutchRegion;


	private boolean updateDataState;
	private boolean redrawDataState;
	private Long startTime;



	/**
	 * The constructor
	 */
	public Activator() {
		plugin = this;
		pcs = new PropertyChangeSupport(plugin);
		activeSensors.add(singleActiveSensor);
	}

	/*
	 * Close and trash current serial port, then reopen new one
	 */
	public void reopenInputPort(String protocol, String portName) {
		saveTimers();
		pauseAll();
		
		//release all resources we might have held before
		if(inputPort!=null) {
			if(inputPort instanceof OSCPortIn) {
				if (((OSCPortIn)inputPort).isListening())
					((OSCPortIn)inputPort).stopListening(); /*close alone not sufficient*/
				
			};
			inputPort.close();
		}
		
		if(protocol.equals(PreferenceConstants.P_INPUT_OSCSERIAL)) {
			System.out.println("Opening OSC Serial port at "+portName);
			inputPort =  (OSCMarshalIn) new OSCSerialPort(portName,getPreferenceStore().getInt(PreferenceConstants.P_COMSPEED));
		} else if  (protocol.equals(PreferenceConstants.P_INPUT_WITILTSERIAL)) {
			System.out.println("Opening WiTilt serial port at "+portName);
			inputPort = (OSCMarshalIn) new WiTiltSerialPort(portName);//because DtoolsSerialPort is abstract, need runtime check
			// and initialize it
			((WiTiltSerialPort)inputPort).initWiTilt();
		} else if (protocol.equals(PreferenceConstants.P_INPUT_UDP)){
			int portNumber = getPreferenceStore().getInt(PreferenceConstants.P_UDP_INPUT_PORT);
			System.out.println("Listening for UDP input on port "+portNumber);
			
			try {
				inputPort = new OSCPortIn(portNumber);
				((OSCPortIn)inputPort).startListening();
			} catch (Exception e) {
				System.err.println("Could not open UDP input port");
				inputPort= null;
			}
		}  else if (protocol.equals(PreferenceConstants.P_INPUT_MAKEUDP)){
			//MAKE CONTROLLER
			try {
				inputPort = new MakeControllerOSCPort(
						getPreferenceStore().getInt(PreferenceConstants.P_UDP_INPUT_PORT), 
						InetAddress.getByName(getPreferenceStore().getString(PreferenceConstants.P_UDP_OUTPUT_IPADDRESS)),
						getPreferenceStore().getInt(PreferenceConstants.P_UDP_OUTPUT_PORT),
						100 //SLOW so firmware doesn't lock up
						);
			} catch (UnknownHostException e) {
				inputPort = null;
				System.err.println("Error when trying to create Make Controller interface.");
				e.printStackTrace();
			}
		}
		else {
			System.err.println("Activator: Don't know what kind of serial port to create.");
			inputPort=null;
		}
		
		restoreTimers();
	}
	
	public void registerInputListeners() {
		if(inputPort==null) {
			System.err.println("could not register input listeners.");
			return;
		}
		//listen to data messages
		for (int i=0; i<ExemplarConstants.NUM_SENSORS; i++) {
			String oscAddress = "/adc".concat(Integer.toString(i));
			inputPort.addListener(oscAddress, updateData);
		}
		//listen to foot pedal clutch events
		inputPort.addListener(ExemplarConstants.OSC_CLUTCH_ADDRESS,updateData);

	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.ui.plugin.AbstractUIPlugin#start(org.osgi.framework.BundleContext)
	 */
	public void start(BundleContext context) throws Exception {
		super.start(context);
		
		
		pcs.addPropertyChangeListener("/adc0", this);
		pcs.addPropertyChangeListener(ExemplarConstants.PROP_CLUTCH,this);
		
		
		reopenInputPort( getPreferenceStore().getString(PreferenceConstants.P_INPUT_PROTOCOL), 
						  getPreferenceStore().getString(PreferenceConstants.P_COMPORT));
		
		
		//create listeners for each attached hardware sensor
		//and create Sensor data models that store the entire sensor signal over time
		for (int i=0; i<ExemplarConstants.NUM_SENSORS; i++) {
			String oscAddress = "/adc".concat(Integer.toString(i));
			//serialPort.addListener(oscAddress, updateData);
			sensors.put(oscAddress,new Sensor(oscAddress));
			
			
			// create some sample events to test functionality
			EventSet es = new EventSet(oscAddress);
			
			
			eventSets.put(oscAddress, es);
			List<String> tempList = new ArrayList<String>();
			tempList.add(oscAddress);
			eventMultiSets.put(tempList, es);
			
			
		}
		
	
		
		// create adapters that generate output:
		//OscOutput oscAdapter = new OscOutput();
		if(PreferenceConstants.P_OUTPUT_FLASH.equals(getPreferenceStore().getString(PreferenceConstants.P_OUTPUT_PROTOCOL))) {
			Runnable flashServer = new FlashServer(getPreferenceStore().getInt(PreferenceConstants.P_TCP_OUTPUT_PORT));
			Thread t = new Thread(flashServer);
			t.start();
			final FlashOutput flashOutput = new FlashOutput();
			System.out.println("Output will be XML messages over TCP; Flash server is running on port "
					+getPreferenceStore().getInt(PreferenceConstants.P_TCP_OUTPUT_PORT));
			
					
		} else if (PreferenceConstants.P_OUTPUT_OSCUDP.equals(getPreferenceStore().getString(PreferenceConstants.P_OUTPUT_PROTOCOL))) {
			final OscOutput oscAdapter = new OscOutput();
			System.out.println("Output will be OSC messages over UDP to "
					+getPreferenceStore().getString(PreferenceConstants.P_UDP_OUTPUT_IPADDRESS)
					+":"
					+getPreferenceStore().getInt(PreferenceConstants.P_UDP_OUTPUT_PORT));
		} else {
			System.err.println("Activator: No valid output protocol found in configuration. Exemplar will not output events.");
		}
		

		registerInputListeners();
		
		//timers for redrawing screen and updating sensor data
		//start these at the end so we don't go out of sync on different sensor objects
		timer = new Timer();
		startTime = System.currentTimeMillis();
		timer.scheduleAtFixedRate(updateData, delay, fastPeriod);
		timer.scheduleAtFixedRate(redrawData, delay+5, slowPeriod); 

	}

	/*
	 * (non-Javadoc)
	 * @see org.eclipse.ui.plugin.AbstractUIPlugin#stop(org.osgi.framework.BundleContext)
	 */
	public void stop(BundleContext context) throws Exception {
		System.out.println("Activator.stop()");
		timer.cancel();
		plugin = null;
		super.stop(context);
		
	}

	/**
	 * Returns the shared instance
	 *
	 * @return the shared instance
	 */
	public static Activator getDefault() {
		return plugin;
	}

	/**
	 * Returns an image descriptor for the image file at the given
	 * plug-in relative path
	 *
	 * @param path the path
	 * @return the image descriptor
	 */
	public static ImageDescriptor getImageDescriptor(String path) {
		return imageDescriptorFromPlugin(PLUGIN_ID, path);
	}
	public Sensor getSensorForOscAddress(String address) {
		return sensors.get(address);
	}
	public EventSet getEventSetForOscAddress(String address) {
		return eventSets.get(address);
	}

	public List<String> getActiveSensors() {
		return activeSensors;
	}

	public void addToActiveSensors(String newAddress, boolean removeIfFound) {
		if(!activeSensors.contains(newAddress)) {
			activeSensors.add(newAddress);
			Collections.sort(activeSensors);
			pcs.firePropertyChange(ExemplarConstants.PROP_SENSOR_UPDATE, null, activeSensors);
			
		} else  {
			if(removeIfFound && activeSensors.size() > 1) {
				activeSensors.remove(newAddress);
				Collections.sort(activeSensors);
				pcs.firePropertyChange(ExemplarConstants.PROP_SENSOR_UPDATE, null, activeSensors);
			}
		}
		setSingleActiveSensor(newAddress);
	}

	public void replaceActiveSensors(String newAddress) {
		activeSensors.clear();
		activeSensors.add(newAddress);
		Collections.sort(activeSensors);
		pcs.firePropertyChange(ExemplarConstants.PROP_SENSOR_UPDATE, null, activeSensors);
		setSingleActiveSensor(newAddress);
	}
	
	public void replaceActiveSensors(List<String> newAddresses) {
		activeSensors.clear();
		activeSensors.addAll(newAddresses);
		Collections.sort(activeSensors);
		pcs.firePropertyChange(ExemplarConstants.PROP_SENSOR_UPDATE, null, activeSensors);
	}

	public void propertyChange(PropertyChangeEvent evt) {
		if(evt.getPropertyName().equals("/adc0")) {
			//currentTime ++;
			if(clutchRegion!=null) {
				//clutchRegion.setEndTime(currentTime);
				clutchRegion.setEndTime(sensors.get("/adc0").getBufferSize());
			}
		}
		else if (evt.getPropertyName().equals(ExemplarConstants.PROP_CLUTCH)) {
			Event activeEvent = getEventSetForOscAddressList(getActiveSensors()).getActiveEvent();
			handleClutch(activeEvent,(((Integer)evt.getNewValue())==1));
			
		}
	}
	
	public void handleClutch(Event evt, boolean turnOn) {
		if(evt!=null) {
			if(turnOn) {
				//start a new region, keep growing it
				clutchRegion = new EventRegion(Activator.getDefault().getCurrentTime(),Activator.getDefault().getCurrentTime()+1,null);
				clutchRegion.setOrigin(EventRegionOrigin.USER_CREATED);
				evt.add(clutchRegion);
			} else {
				//stop the new region if we had one.
				if(clutchRegion!=null) {
					if(getPreferenceStore().getString(PreferenceConstants.P_ADJUSTREGION).equals(PreferenceConstants.P_ADJUSTON)) {
						evt.adjustRegion(clutchRegion);
					}
					evt.recalculate(true);
					clutchRegion = null;
				}
			}
		}
	}
	
	public int getCurrentTime() {
		return sensors.get("/adc0").getBufferSize();
	}
	public Set<String> getSensorNames()
	{
		return sensors.keySet();
	}
	public EventSet getEventSetForOscAddressList(List<String> addresses) {
		return eventMultiSets.get(addresses);
	}
	public EventSet createEventSetForOscAddressList(List<String> addresses) {
		//check that we're not creating duplicates
		if(eventMultiSets.containsKey(addresses))
			return eventMultiSets.get(addresses);
		
		EventSet es = new EventSet(addresses);
		eventMultiSets.put(addresses, es);
		return es;
	}
	
	

	public Sensor getSingleActiveSensor() {
		//return getSensorForOscAddress(getActiveSensors().get(0));
		return getSensorForOscAddress(singleActiveSensor);
	}
	
	public void setSingleActiveSensor(String newAddress)
	{
		if(activeSensors.contains(newAddress))
			singleActiveSensor = newAddress;
		else
			singleActiveSensor = activeSensors.get(0);
		
		pcs.firePropertyChange(ExemplarConstants.PROP_ACTIVE_SENSOR_UPDATE, null, singleActiveSensor);
	}
	
	public EventSet getActiveEventSet() {
		return getEventSetForOscAddressList(getActiveSensors());
	}
	
	public void eventSetRecalculate() 
	{
		List<Event> eventSet = getActiveEventSet().getEvents();
	
		for(Iterator it = eventSet.iterator(); it.hasNext();)
		{
			((Event) (it.next())).recalculate(true);
		}
	}
	/**
	 * Experimental - not tested.
	 * @see http://java.sun.com/developer/technicalArticles/Programming/serialization/
	 * @param filename
	 */
	public void saveToFile(String filename) {
		//pause the timer tasks (i.e., fire no events)
		saveTimers();
		pauseAll();
		
		if(filename==null || filename.equals("")) {
			filename = "exemplar-session.exm";
			
		}
		//FileOutputStream fos = null;
		FileWriter fw = null;
		ObjectOutputStream out = null;
		XStream xstream = new XStream();
		try
		{
			//fos = new FileOutputStream(filename);
			fw = new FileWriter(filename);
			out = xstream.createObjectOutputStream(fw);
			//out = new ObjectOutputStream(fos);
			out.writeObject(sensors);
			out.writeObject(eventSets);
			out.writeObject(eventMultiSets);
			out.close();
			fw.close();
		}
		catch(IOException ex)
		{
			ex.printStackTrace();
		}
		//turn timer tasks back on
		restoreTimers();
	}
	
	
	@SuppressWarnings("unchecked")
	public void loadFromFile(String filename) {
		saveTimers();
		pauseAll();
		//go through all sensors, highlighting one at a time
		//WHY is this here?
		for (int i=ExemplarConstants.NUM_SENSORS-1; i>=0; i--) {
			String oscAddress = "/adc".concat(Integer.toString(i));
			replaceActiveSensors(oscAddress);
			pcs.firePropertyChange(ExemplarConstants.PROP_SENSOR_UPDATE, null, getActiveSensors());
		}

		
		XStream xstream = new XStream();
		if(filename==null || filename.equals("")) {
			filename = "exemplar-session.exm";
			
		}
		FileReader fr = null;
		//FileInputStream fis = null;
		ObjectInputStream in = null;
		try
		{
			//fis = new FileInputStream(filename);
			fr = new FileReader(filename);
			in = xstream.createObjectInputStream(fr);
			//in = new ObjectInputStream(fis);
			sensors = (Map<String, Sensor>) in.readObject();
			eventSets = (Map<String, EventSet>) in.readObject();
			eventMultiSets = (Map<List<String>, EventSet>) in.readObject();
			in.close();
		}
		
		catch(IOException ex)
		{
			ex.printStackTrace();
		}
		catch(ClassNotFoundException ex)
		{
			ex.printStackTrace();
		}
		//reset current time to last time recorded in file
		 
		//update the UI
		//replaceActiveSensors("/adc0");
		//pcs.firePropertyChange(ExemplarConstants.PROP_SENSOR_UPDATE, null, getActiveSensors());
		
		//update all single sensors
		for (int i=ExemplarConstants.NUM_SENSORS-1; i>=0; i--) {
			String oscAddress = "/adc".concat(Integer.toString(i));
			replaceActiveSensors(oscAddress);
			pcs.firePropertyChange(ExemplarConstants.PROP_SENSOR_UPDATE, null, getActiveSensors());
		}
			
			
		//turn timer tasks back on
		restoreTimers();
	}

	public void pauseInput() {
		updateData.pause();
	}
	public void resumeInput() {
		updateData.resume();
	}

	public void clearAllSensorData() {
		saveTimers();
		pauseAll();
		
		//clear all recorded sensor data
		for(Sensor s:sensors.values()) {
			s.clear();
		}
		//clear all event regions, since their times are meaningless now
		for(EventSet es: eventSets.values()) {
			for(Event e: es.events) {
				e.clearRegions();
			}
		}
		for(EventSet es: eventMultiSets.values()) {
			for(Event e: es.events) {
				e.clearRegions();
			}
		}
		//restart listening to input
		restoreTimers();
		
	}
	/**
	 * Get rid of all user defined events
	 *
	 */
	public void clearAllEvents() {
		saveTimers();
		pauseAll();

		
		// iterate over all eventSets, and create new empty eventSet and swap it out.
		Activator.getDefault().getWorkbench().getDisplay().asyncExec (new Runnable () {
			public void run () {
				for(Map.Entry<List<String>, EventSet>entry: eventMultiSets.entrySet()) {
					List<String> oscAddresses = entry.getValue().getOscAddresses();
					
				
					
					//	create a new empty one
					EventSet newSet = new EventSet(oscAddresses);
					
					//	swap them in gui
					pcs.firePropertyChange(ExemplarConstants.PROP_REPLACE_EVENT_SET_MODEL,oscAddresses,newSet);
					//	write new empty set back into data structure
					entry.setValue(newSet);
					if(oscAddresses.size()==1) {
						eventSets.remove(oscAddresses.get(0));
						eventSets.put(oscAddresses.get(0), newSet);
					}
				}
			}});
		//restart listening to input
		restoreTimers();
	}
	
	public void pauseAll() {
		updateData.pause();
		redrawData.pause();
		
	}
	public void resumeAll() {
		updateData.resume();
		redrawData.resume();
		
	}

	public void saveTimers() {
		updateDataState = updateData.isPaused();
		redrawDataState = redrawData.isPaused();
		
	}

	public void restoreTimers() {
		
		updateData.setPaused(updateDataState);
		redrawData.setPaused(redrawDataState);
		
	}

	public Collection<EventSet> getEventSets() {
		return eventMultiSets.values();
	}

	public void saveRawSensorDataToFile(String sensor, String filename) {
		
		saveTimers();
		pauseAll();
		
		if(filename==null || filename.equals("")) {
			filename = "exemplar-save.exm";
			
		}
		//FileOutputStream fos = null;
		FileWriter fw = null;
		ObjectOutputStream out = null;
		XStream xstream = new XStream();
		try
		{
			//fos = new FileOutputStream(filename);
			fw = new FileWriter(filename);
			out = xstream.createObjectOutputStream(fw);
			//out = new ObjectOutputStream(fos);
			out.writeObject(sensor);
			out.writeObject(startTime);
			out.writeObject(sensors.get(sensor).getRecordedData());
			out.close();
			fw.close();
		}
		catch(IOException ex)
		{
			ex.printStackTrace();
		}
		//turn timer tasks back on
		restoreTimers();
	}

	public Event getEventForOutputName(String name) {
//		clear all event regions, since their times are meaningless now
		for(EventSet es: eventSets.values()) {
			for(Event e: es.events) {
				if(name.equals(e.key))
					return e;
			}
		}
		for(EventSet es: eventMultiSets.values()) {
			for(Event e: es.events) {
				if(name.equals(e.key))
					return e;
			}
		}
		return null;
	}
	/*
	public void sendMessageToHardware(String address, Integer argument) {
		OSCMessage msg = new OSCMessage();
		msg.setAddress(address);
		msg.addArgument(argument);
		sendMessageToHardware(msg);
	}
	
	public void sendMessageToHardware(OSCMessage msg) {
		try {
			serialPort.send(msg);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
    */
	
	public void initializeWiTilt() {
		if(inputPort instanceof WiTiltSerialPort) {
			((WiTiltSerialPort)inputPort).initWiTilt();
		}
		
	}
	public long getSamplingPeriodMs() {
		return fastPeriod;
	}

}