package edu.stanford.hci.exemplar;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
/**
 * Exemplar
 * Copyright (c) 2004-2007, Stanford University
 * All rights reserved.
 *
 * <p>
 * This software is distributed under the <a href="http://hci.stanford.edu/research/copyright.txt">
 * BSD License</a>.
 * </p>
 * 
 * 
 * @author <a href="http://bjoern.org/">Bjoern Hartmann</a> ( bjoern(AT)stanford.edu )
 * @author Leith Abdulla
 */
public class PatternEventEvaluator implements EventEvaluator {

	private static final long serialVersionUID = 1L;

	private float maxDistance, median; //maximum DTW distance for a match
	int updateInterval, updateWidthInterval;
	private ArrayList<Float> pattern = new ArrayList<Float>();
	//private ArrayList<Float> error = new ArrayList<Float>();
	private Map<Integer, Float> error = new HashMap<Integer, Float>();
	private enum patternMode {NONE, INIT, DONE};
	private patternMode mode = patternMode.NONE;
	private float lastDistance;
	Event event;
	private int startTime;
	String oscAddress;
	
	public PatternEventEvaluator(Event evt, String oscAddress) {
		super();
		updateInterval=2;
		updateWidthInterval = 2;
		event = evt;
		this.oscAddress = oscAddress;
	}


	public boolean evaluate(int curTime, int width, boolean fireEvents) {
				
		if(pattern.size() > 0)
		{
			Float distance = dtw(curTime, width);
			error.put(curTime, Math.abs(distance));
			
			//complicated logic cut out here
			
			// if we have a match:
			if(Math.abs(distance) < maxDistance) {
				//create a region that highlights the range we just calculated over
				EventRegion r = new EventRegion(curTime-width,curTime, oscAddress);
				r.setOrigin(EventRegionOrigin.SYSTEM_CREATED);
				//add it and fire output
				event.add(r);
				if(fireEvents) {
					event.evaluatorNotify(this,true);
					//event.evaluatorNotify(this,false);
				}
				return true;
			} else {
				return false;
			}
		}
		return false;
	}

	
	 private  float dtw(int curtime, int width) {
		 //s (length n) - sensor values
		 // t (length m) - pattern we're matching;
		 //d is the distance matrix - we create
		 //translated into Java from http://en.wikipedia.org/wiki/Dynamic_time_warping
		 Sensor s = Activator.getDefault().getSensorForOscAddress(oscAddress);
		 int n = width;
		 int m = pattern.size();
		 
		 float[][] DTW = new float[n+1][m+1];
		 for(int i=0; i<=m; i++) {
			 DTW[0][i] = Float.MAX_VALUE;
		 }
		 for(int i=0; i<=n; i++) {
			 DTW[i][0] = Float.MAX_VALUE;
		 }
	     DTW[0][0] = 0;
	     for(int i=1; i<=n; i++) {
	    	 for (int j=1; j<=m; j++) {
	    		 float cost = Math.abs(s.getFilteredValueforTime(curtime-width+(i-1))-pattern.get(j-1));
	    		 DTW[i][j]=Math.min(DTW[i-1][j] + cost, Math.min(DTW[i][j-1]+cost,DTW[i-1][j-1]+cost));
	    	 }
	     }
		 return DTW[n][m];
	 }

	/**
	 * 
	 * @param curTime, width 
	 * Returns unweighted distance between the 2 signals <values>
	 * Lower indicates a closer match
	 * @return
	 */
	private Float dtw2(int curTime, int width) 
	{
		Sensor s = Activator.getDefault().getSensorForOscAddress(oscAddress);
		int len1 = width; //Arr1.length;
		int len2 = pattern.size(); //Arr2.length;
		

		
		float[][] d = new float[len1][len2];
		float[][] D = new float[len1][len2];
		//float[][] Distance = new float[len1][len2];
//		 Initialize the array d to 0;
		for ( int i = 0; i < len1; i++) 
			for (int j = 0 ; j < len2; j++) {
				//d[i][j] = Arr1[i] - Arr2[j]; // Calculate Local Distances for d;
				d[i][j] = s.getFilteredValueforTime(curTime - len1 + i) - pattern.get(j);
				D[i][j] = 0;
			}
		D[0][0] = d[0][0];	// Global Distance
		
		for(int i = 1; i < len1; i++)
			D[i][0] =  d[i][0] + D[i-1][0];
		
		for(int j = 1; j < len2; j++)
			D[0][j] = d[0][j] + D[0][j-1];
		
		for ( int i = 1; i < len1; i++) 
			for (int j = 1 ; j < len2; j++) {
				D[i][j] = d[i][j] + Math.min(D[i-1][j-1], Math.min(D[i-1][j],D[i][j-1] ));
			}
		// Start Tracing the path from the reverse direction
		int n = len1 - 1;
		int m = len2 - 1;
		Float values;
		values = new Float(0.0);
		
		while(m + n != 2) {
			if(n == 1)
				m--;
			else if(m == 1)
				n--;
			else {
				if(D[n-1][m] < D[n][m-1]) {
					if(D[n-1][m] < D[n-1][m-1]) {
						values = D[n-1][m];
						n--;
					}
					else {
						values = D[n-1][m-1];
						n--;
						m--;
					}
				}
				else {
					if(D[n][m-1] < D[n-1][m-1]) {
						values = D[n][m-1];
						m--;
					}
					else {
						values = D[--n][--m];
					}
				}
			}			
		}
		return values;
	}

	public int getUpdateInterval() {
		return updateInterval;
	}

	public int getUpdateWidthInterval() {
		return updateWidthInterval;
	}
	
	public float getThreshold()
	{
		return maxDistance;
	}
	
	public void setThreshold(float t)
	{
		maxDistance = t;
	}
	
	public float getMedian()
	{
		return median;
	}
	
	public void learn(List<EventRegion> examples) {
	
		if(examples.size() > 0)
		{
			Sensor s = Activator.getDefault().getSensorForOscAddress(oscAddress);
			float n, last = 0, m = 0;
			pattern.clear();
			error.clear();
			
			
			/* for(int i = examples.get(0).startTime; i < examples.get(0).endTime; i++)
			{
				n = s.getFilteredValueforTime(i);
				m += n - last;
				pattern.add(n);
				last = n;
			}*/
			
			/* old patterns that compares the pattern to a flat line*/
			for(int i = examples.get(0).startTime; i < examples.get(0).endTime; i++)
			{
				n = s.getFilteredValueforTime(i);
				m += n;
				pattern.add(n);
			}
			
			
			//updateInterval = (int)(pattern.size() * 0.6);
			//updateWidthInterval = (int)(pattern.size() * 1.4);
			updateInterval = 2;
			updateWidthInterval = 2;
			
			//find good automatic threshold and median line
			//heuristic: what immediately preceded should be OUT
			//running the dtw on itself should be IN
			int width = examples.get(0).endTime- examples.get(0).startTime;
			float out = dtw( examples.get(0).startTime, width); //this is segment preceding
			float in = dtw( examples.get(0).endTime-5, width);
			median = out;
			maxDistance = Math.max(out/2,in+(out-in)/2.f); 
			
		}
	}

	public Map<Integer, Float> getPatternError()
	{
		return error;
	}
	
	public void clear()
	{
		pattern.clear();
		error.clear();
	}


	public EventType getType() {
		return EventType.PATTERN;
	}
//	 first attempt at writing pseudocode to export evaluator outside exemplar
	// haven't added hysteresis or timeouts yet
	public String getPseudocode() {
		String s = "";
		s+="bool evaluate() {\n";
		s+="\t//note: patterns not yet supported.\n";
		s+="\t\treturn false;\n";
		s+="}\n";
		return s;
	}

	/**
	 * Get avg power in signal - move end in until it goes beyond a threshold (i.e., action occurs)
	 * 
	 */
	public boolean adjustRegion(EventRegion clutchRegion) {
		Sensor sensor = Activator.getDefault().getSensorForOscAddress(oscAddress);
		// sStart----rStart=====rEndsEnd
		//r = region
		int rStart = clutchRegion.startTime;
		int rEnd = clutchRegion.endTime;
		//s = search
		
		float pwr;
		float minPwr = Float.MAX_VALUE;
		float maxPwr = Float.MIN_VALUE;
		//ITERATE OVER MARKED REGION, moving min/maxs in
		final int SEARCH_SIZE = 5;
		for(int t=rStart; t<(rEnd-SEARCH_SIZE); t++) {
			/* calculate min, max power in signal */
			pwr = 0;
			for(int i=0; i<SEARCH_SIZE; i++) {
				pwr+= Math.abs(sensor.getFilteredValueforTime(t+i+1)-sensor.getFilteredValueforTime(t+i));
			}
			if(pwr>maxPwr)
				maxPwr = pwr;
			if(pwr<minPwr)
				minPwr = pwr;
		}
		float cutoff = minPwr + (maxPwr-minPwr)/10; //HACK: hardcoded
		int newStartTime=rStart;
		int newEndTime=rEnd;
		
		//go forward and move start of region in
		for(int t=rStart; t<(rEnd-SEARCH_SIZE); t++) {
			/* power in signal */
			pwr = 0;
			for(int i=0; i<SEARCH_SIZE; i++) {
				pwr+= Math.abs(sensor.getFilteredValueforTime(t+i+1)-sensor.getFilteredValueforTime(t+i));
			}
			if(pwr>cutoff) {
				newStartTime=t;
				break;
			}
		}
		//go backwards
		for(int t=rEnd-(SEARCH_SIZE); t>=rStart; t--) {
			/* power in signal */
			pwr = 0;
			for(int i=0; i<SEARCH_SIZE; i++) {
				pwr+= Math.abs(sensor.getFilteredValueforTime(t+i+1)-sensor.getFilteredValueforTime(t+i));
			}
			if(pwr>cutoff) {
				newEndTime=t;
				break;
			}
		}
		clutchRegion.setStartTime(newStartTime);
		clutchRegion.setEndTime(newEndTime);
		//System.err.println("PatternEventEvaluator.adjustRegion ("+rStart+","+rEnd+") -> ("+newStartTime+","+newEndTime+")");
		//System.err.println("Min:"+minPwr+"; Max: "+maxPwr+"; cutoff: "+cutoff);
		return true;
	}


	public int getPatternSize() {
		return pattern.size();
	}
}
