package edu.stanford.hci.exemplar.flashserver;

import java.io.IOException;
import java.io.PrintWriter;
import java.net.ServerSocket;
import java.net.Socket;

import com.illposed.osc.OSCMessage;

import edu.stanford.hci.dtools.serial.OSCSerialPort;

/**
 * XmlSocketServer to talk to Flash from Java. 
 * For the moment, only one direction (Java->Flash) is implemented.
 * Flash->Java to follow later on.
 * networking code is adapted from : http://java.sun.com/docs/books/tutorial/networking/sockets/clientServer.html
 *
 * Copyright (c) 2004-2007, Stanford University
 * All rights reserved.
 *
 * <p>
 * This software is distributed under the <a href="http://hci.stanford.edu/research/copyright.txt">
 * BSD License</a>.
 * </p>
 * 
 * 
 * @author <a href="http://bjoern.org/">Bjoern Hartmann</a> ( bjoern(AT)stanford.edu )
 * @author Leith Abdulla
 * TODO: don't call System.exit(-1).
 */
public class FlashServer implements Runnable{
	private static final String END_OF_LINE="\n\000"; //string termination for XML fragments
	private int portNum=0;
	ServerSocket serverSocket=null;
	Socket clientSocket=null;
	PrintWriter out=null;
	private OSCSerialPort dtoolsPort; 
	boolean listening = true;
	
	private long lastTime=0;
	static FlashServer uniqueInstance;

	/**
	 * Starts a server and waits for a client to connect (blocking); exits on any errors.
	 * @param comPort 
	 * @param dtoolsPort the local TCP dtoolsPort the Server will listen on
	 */
	public FlashServer(int tcpPort) {
		uniqueInstance=this;
		portNum=tcpPort;
	}
	
	//Singleton pattern
	public static FlashServer instance() {
		return uniqueInstance;
	}
	/**
	 * Forward a message to flash. Format of the XML fragment is:
	 * <dtoolsevent address="eventaddress" value="eventvalue" />
	 * 
	 * @param address the OSC address of the hardware component generating the event, e.g, "/btn2" or "/in1"
	 * @param argument the value of the event, e.g. 1 for a button press, 0 for a button release, a number between 0 and 255 for a slider...
	 */
	public void sendMessageToFlash(String msg) {
		long now = System.currentTimeMillis();
		//System.err.println("sendMessageToFlash: "+msg+" delta: "+(now-lastTime));
		lastTime=now;
		String outputLine = msg+END_OF_LINE;
		if(out!=null) {
		try {
			out.write(outputLine);
			if( out.checkError() ) {
				System.out.println("FlashServer: Client caused a write error. Ignoring.");
				
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		} else {
			System.err.println("FlashServer: Flash port not open yet. Not sending msg.");
		}
	}

	public void sendMessageToExemplar(String address, Object argument) {
		try {
			OSCMessage msg = new OSCMessage(address);
			msg.addArgument(argument);
			dtoolsPort.send(msg);
		} catch (IOException e) {
			System.err.println("FlashServer: Could not send message "+address+","+argument+" to Hardware.");
		}
	}
	
	
	public synchronized void quit() {
		listening = false;
		try {
			serverSocket.close();
		} catch (IOException e1) {
			e1.printStackTrace();
		}
	}

	public void run() {
		// create an xml server - blocks until Flash client connects
		// try to create server
		try{
			serverSocket = new ServerSocket(portNum);
			System.out.println("FlashServer: Created ServerSocket; listening for clients on port "+portNum+"...");
		} catch (IOException e) {
			System.err.println("FlashServer: Could not listen on port "+portNum);
			System.exit(-1);
		}
		while(listening) {
			// wait for client to connect (blocking)
			try {
				clientSocket = serverSocket.accept();
				System.out.println("FlashServer: Accepted connection from "+clientSocket.getInetAddress()+", port "+clientSocket.getPort());
			} catch (IOException e) {
				System.err.println("FlashServer: Accept failed on port "+portNum);
				System.exit(-1);
			}

			// create input reader and output writer
			try {
				out = new PrintWriter(
						clientSocket.getOutputStream(), true);
				Thread t = new Thread(new FlashListener(clientSocket, this));
				t.start();
			} catch (IOException e) {
				System.err.println("FlashServer: Could not get input or output stream");
				System.exit(-1);
			}
		}
		//if we get this far, we're now ready to communicate with flash. great.
		
	}
}

