package edu.stanford.hci.exemplar;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.swt.graphics.RGB;

/**
 * Base class for Events
 *
 * Exemplar
 * Copyright (c) 2004-2007, Stanford University
 * All rights reserved.
 *
 * <p>
 * This software is distributed under the <a href="http://hci.stanford.edu/research/copyright.txt">
 * BSD License</a>.
 * </p>
 * 
 * 
 * @author <a href="http://bjoern.org/">Bjoern Hartmann</a> ( bjoern(AT)stanford.edu )
 * @author Leith Abdulla
 */
public abstract class Event implements PropertyChangeListener, Serializable {
	
	private final List<EventRegion> regions = new ArrayList<EventRegion>();
	protected transient List<EventObserver> observers = new ArrayList<EventObserver>();
	protected EventType type = EventType.THRESHOLD;
	protected String name;
	private static final String defaultName = "MyEvent";
	protected String key;
	protected RGB color;
	protected EventSet parentSet;
	protected boolean active = true;
	protected boolean doOutput = false;
	
	private static int count=0;
	public Event(EventSet parent) {
		super();
		this.parentSet=parent;
		this.color = new RGB(255,0,0);
		name = defaultName;
		//key="/out"+Integer.toString(count++);
		key=null;
		// HACK
		Activator.getDefault().pcs.addPropertyChangeListener("/adc0",this);
		
	}
	public Event(Event copy) {
		this.regions.addAll(copy.regions);
		this.observers.addAll(copy.observers);
		this.type = copy.type;
		this.name = copy.name;
		this.key = copy.key;
		this.color = copy.color;
		this.parentSet=  copy.parentSet;
		this.active = copy.active;
		
	}

	public RGB getColor() {
		return color;
	}
	public void setColor(RGB color) {
		this.color = color;
		notifyObservers();
	}
	public synchronized List<EventRegion> getRegions() {
			return regions;
	}

	public synchronized boolean add(EventRegion arg0) {
		boolean b = regions.add(arg0);
		if(arg0.origin.equals(EventRegionOrigin.USER_CREATED)) {
			
					Activator.getDefault().pcs.firePropertyChange(ExemplarConstants.PROP_ADD_EVENT_REGION, this, arg0);
		
		}
			
		return b;
	}

	public synchronized boolean remove(Object o) {
			return regions.remove(o);
	}
	public abstract EventType getType(String string);
	
	public abstract void setType(EventType type, String string);

	public void attach(EventObserver observer) {
		observers.add(observer);
	}
	public void detach(EventObserver observer) {
		observers.remove(observer);
	}
	protected void notifyObservers() {
		for(EventObserver o: observers) {
			o.Update();
		}
	}
	public String getName() {
		return name;
	}
	public String getKey() {
		return key;
	}
	public void setKey(String key) {
		this.key = key;
		doOutput = (key!=null);
		Activator.getDefault().pcs.firePropertyChange(ExemplarConstants.PROP_RENAME_EVENT_OUTPUT, null, this);
		if(key!=null) {
			parentSet.setModified(true);
		}
		notifyObservers();
	}
	public void setName(String name) {
		this.name = name;
		Activator.getDefault().pcs.firePropertyChange(ExemplarConstants.PROP_RENAME_EVENT, null, this);
		if(!defaultName.equals(name)) {
			parentSet.setModified(true);
		}
		notifyObservers();
	}
	public boolean isActive() {
		return active;
	}
	//called from parent
	public void setActive(boolean active) {
		this.active = active;
		notifyObservers();
	}
	//called from Widget
	public void makeActive() {
		parentSet.setActiveEvent(this);
	}
	/**
	 * Compute whether this event should be currently active or not
	 */
	public abstract void update();
	
	
	
	/**
	 * Recalculate regions where current event is true for the entire history
	 * of the event's sensor data. Call this method when adding a new user region,
	 * deleting a user region, or changing event type.
	 */
	public abstract void recalculate(boolean relearn);	
	
	/**
	 * Adjust boundaries to minimize spurious matches
	 * TODO: implement
	 * @param clutchRegion 
	 */
	public abstract void adjustRegion(EventRegion clutchRegion);
	
	public abstract EventEvaluator getEvaluator(String OscAddress);
	
	public EventSet getEventSet() {
		return parentSet;
	}
	
	public void propertyChange(PropertyChangeEvent evt) {
		//if ...equals... (skip b/c we're only listening to one thing);
		
		update();
	}
	/**
	 * Evaluator uses this function to notify event of on/off state changes
	 */
	public abstract void evaluatorNotify(EventEvaluator evaluator, boolean newState);
	
	/**
	 * Write object out to file
	 * @param out
	 * @throws IOException
	 */
	private void writeObject(ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
	}
	
	/**
	 * Read object back in from file and reregister with the propertychangesupport.
	 * 	@param in
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		Activator.getDefault().pcs.addPropertyChangeListener("/adc0",this);
		//observers get refreshed via the event set
		observers = new ArrayList<EventObserver>();
	}
	/**
	 * we're about to be deleted, so stop listening and clear all our children
	 *
	 */
	public void dispose() {
		Activator.getDefault().pcs.removePropertyChangeListener("/adc0",this);
		active=false;
		observers.clear();
		regions.clear();
		
	}
	public void clearRegions() {
		//first fire a remove event for all user regions
		for(EventRegion r:regions) {
			if(r.origin.equals(EventRegionOrigin.USER_CREATED)) {
				Activator.getDefault().pcs.firePropertyChange(ExemplarConstants.PROP_REMOVE_EVENT_REGION, this, r);
			}
		}
		//now we're safe to clear
		regions.clear();
	}
	public void removeRegion(EventRegion region) {
		Activator.getDefault().pcs.firePropertyChange(ExemplarConstants.PROP_REMOVE_EVENT_REGION, this, region);
		regions.remove(region);
	}
	public abstract String getPseudocode();

	public boolean doesOutput() {
		return doOutput;
	}
	public void setOutput(boolean doOutput) {
		this.doOutput = doOutput;
	}

}
