package edu.stanford.hci.exemplar;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
/**
 * Exemplar
 * Copyright (c) 2004-2007, Stanford University
 * All rights reserved.
 *
 * <p>
 * This software is distributed under the <a href="http://hci.stanford.edu/research/copyright.txt">
 * BSD License</a>.
 * </p>
 * 
 * 
 * @author <a href="http://bjoern.org/">Bjoern Hartmann</a> ( bjoern(AT)stanford.edu )
 * @author Leith Abdulla
 */
public class MultiSensorEvent extends Event {
	
	private static final long serialVersionUID = 1L;
	protected Map<String,EventEvaluator> oscToEvalMap = new HashMap<String,EventEvaluator>(); 
	protected Map<EventEvaluator,Boolean> evalMap = new HashMap<EventEvaluator,Boolean>();
	protected Map<EventEvaluator,Integer> lastTimeMap = new HashMap<EventEvaluator,Integer>();
	private boolean haveFiredEvent=false;
	public MultiSensorEvent(EventSet parent) {
		super(parent);
		for(String address: parent.getOscAddresses()) {
			EventEvaluator eval = new ThresholdEventEvaluator(this,address);
			oscToEvalMap.put(address, eval);
			evalMap.put(eval, false);
			lastTimeMap.put(eval, 0);
		}
	}


	@Override
	public void recalculate(boolean relearn) {
		Activator.getDefault().saveTimers();
		Activator.getDefault().pauseAll();
		
				//throw out all system-created regions
				List<EventRegion> keepers = new ArrayList<EventRegion>();
				for(EventRegion r:getRegions()) {
					if (r.getOrigin().equals(EventRegionOrigin.USER_CREATED)) {
						keepers.add(r);
					}
				}
				getRegions().clear();
				getRegions().addAll(keepers);
				//if(keepers.size()>0) {
					for(EventEvaluator evaluator:oscToEvalMap.values())	{	
						
						if(relearn)
							evaluator.learn(keepers);
						//	recreate system regions based on user regions
	
						if(evaluator.getType().equals(EventType.THRESHOLD))
						{
							evaluator.evaluate(Activator.getDefault().getCurrentTime()-1,Activator.getDefault().getCurrentTime()-1,false);
						}
						else if(evaluator.getType().equals(EventType.PATTERN))
						{
							//	don't re-evaluate in pattern mode for now
						}
					
					}
				//}
				//we're probably not in the UI thread, so notify our Widget observers with the async thread switch workaround
				Activator.getDefault().getWorkbench().getDisplay().asyncExec (new Runnable () {
					public void run () {
						notifyObservers();
					}
				});
		Activator.getDefault().restoreTimers();
	}

	@Override
	public void update() {
		for(EventEvaluator evaluator:oscToEvalMap.values())	{	
			//BUG!	need to maintain last time for all sensors
			int lastTime = lastTimeMap.get(evaluator);
			int curTime = Activator.getDefault().getCurrentTime()-2;
			int width = (curTime-lastTime);
			if(width>=evaluator.getUpdateInterval()) {
				
				if(evaluator.evaluate(curTime,width,true))
				{
					lastTimeMap.put(evaluator, curTime);
				}
				else if(width >= evaluator.getUpdateWidthInterval())
				{
					lastTimeMap.put(evaluator, curTime);
				}
						//if anything changed:
						//notifyObservers();
			}
		}
	}

	@Override
	/**
	 * Note: this method only works the intersection of intervals when events are detected overlap
	 * Should be changed so the evaluator passes in times
	 */
	public void evaluatorNotify(EventEvaluator evaluator, boolean newState) {
		System.err.println("evaluatorNotify"+evaluator+" "+newState);
		evalMap.remove(evaluator);
		evalMap.put(evaluator,newState);
		boolean allTrue = true;
		for(Map.Entry<EventEvaluator,Boolean> entry : evalMap.entrySet()) {
			if(!entry.getValue()) {
				allTrue=false;
				break;
			}
		}
		if(allTrue && !haveFiredEvent) {
			if(doOutput)
				Activator.getDefault().pcs.firePropertyChange(ExemplarConstants.PROP_EVENT_ON, null, key);
		    haveFiredEvent = true;
		    return;
		} 
		if(!allTrue && haveFiredEvent) {
			if(doOutput)
				Activator.getDefault().pcs.firePropertyChange(ExemplarConstants.PROP_EVENT_OFF, null, key);
			haveFiredEvent=false;
			return;
		}
	}

	@Override
	public void setType(EventType type, String address)  {
		// if we're not changing event type, do nothing 
		// (i.e., don't trash the evaluator we have that has learned state) 
		if(oscToEvalMap.get(address)!= null && oscToEvalMap.get(address).equals(type))
			return;
		
		evalMap.remove(oscToEvalMap.get(address));
		lastTimeMap.remove(oscToEvalMap.get(address));
		oscToEvalMap.remove(address);
		
		if(type.equals(EventType.THRESHOLD)) {
			EventEvaluator eval = new ThresholdEventEvaluator(this,address);
			oscToEvalMap.put(address, eval);
			evalMap.put(eval, false);
			lastTimeMap.put(eval, 0);
		} else if (type.equals(EventType.PATTERN)) {
			EventEvaluator eval = new PatternEventEvaluator(this,address);
			oscToEvalMap.put(address, eval);
			evalMap.put(eval, false);
			lastTimeMap.put(eval, 0);
		} else {
			
		}
		printMap(oscToEvalMap);
	}

	private void printMap(Map<String, EventEvaluator> oscToEvalMap2) {
		for(Map.Entry<String, EventEvaluator> e:oscToEvalMap2.entrySet()) {
			System.err.println("Key: "+e.getKey()+"; Val: "+e.getValue() );
		}
		System.err.println("-----------");
	}


	@Override
	public EventEvaluator getEvaluator(String OscAddress) {
		return oscToEvalMap.get(OscAddress);
	}

	@Override
	public EventType getType(String string) {
		return(oscToEvalMap.get(string).getType());
	}


	@Override
	public String getPseudocode() {
		return "//getPseudocode for MultiSensorEvent not yet implemented. Sorry.\n\n";
	}


	@Override
	public void adjustRegion(EventRegion clutchRegion) {
		for(EventEvaluator evaluator:oscToEvalMap.values())	{	
			evaluator.adjustRegion(clutchRegion);
		}
	}
}
