package edu.stanford.hci.exemplar;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Vector;

import edu.stanford.hci.exemplar.exensions.FilterExtension;
import edu.stanford.hci.exemplar.exensions.FilterExtensionLoader;
import edu.stanford.hci.exemplar.preferences.PreferenceConstants;
/**
 * Exemplar
 * Copyright (c) 2004-2007, Stanford University
 * All rights reserved.
 *
 * <p>
 * This software is distributed under the <a href="http://hci.stanford.edu/research/copyright.txt">
 * BSD License</a>.
 * </p>
 * 
 * 
 * @author <a href="http://bjoern.org/">Bjoern Hartmann</a> ( bjoern(AT)stanford.edu )
 * @author Leith Abdulla
 * @author Manas Mittal*/
public class Sensor implements PropertyChangeListener, Serializable {
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	/*
	private static final int DIRECTIONTHRESH = 2;
	private static final int MAXMINWIDTH = 5;
	*/
	private Vector<FilterEntry> filterList; // uses FilterType enum
	private final String OSCAddress;
	private ArrayList<Float> recordedData; //BIG
	int width = 4; //default width for smoothing filter
	Float sd = new Float(2.0);
	private Float scaleFactor = 1.0f;
	private float offset=0.f;	
	private transient Map<Integer,Float> cache; //cache for filtered data entries
	private int hit=0,miss=0,count=0;
	private String oscOutputAddress=null;
	private boolean outputOsc;
	private ScaleBaseline scaleBaseline = ScaleBaseline.BOTTOM;
	
	public Sensor(String OSCAddress) {
		super();
		this.OSCAddress=OSCAddress;
		recordedData = new ArrayList<Float>();
		Activator.getDefault().pcs.addPropertyChangeListener(OSCAddress,this);
		filterList = new Vector<FilterEntry> ();
		//add our list of hardcoded filters
		filterList.add(new FilterEntry(FilterType.DIFF,false));
		filterList.add(new FilterEntry(FilterType.SCALE,false));
		filterList.add(new FilterEntry(FilterType.OFFSET,false));
		filterList.add(new FilterEntry(FilterType.SMOOTHE,true)); //HACK: turn smoothing on by default
		//filterList.add(new FilterPair(FilterType.MAP,false));

		//add here: load in user-defined filters
		for(Class cls: FilterExtensionLoader.getInstance().getFilterList()) {
			filterList.add(new FilterEntry(FilterType.USERDEFINED,false,cls));
		}
		//filterList.add(new FilterEntry(FilterType.USERDEFINED,false,CustomFilter1.class));
		//filterList.add(new FilterEntry(FilterType.USERDEFINED,false,CustomFilter2.class));
		initCache();
	}
	
	/**
	 * @see http://javaalmanac.com/egs/java.util/coll_Cache.html
	 *
	 */
	@SuppressWarnings("serial")
	private void initCache() {
	    // Create cache
		
	    final int MAX_ENTRIES = 1200;
	    cache = new LinkedHashMap<Integer,Float>(MAX_ENTRIES+1, .75F, true) {
	        // This method is called just after a new entry has been added
	        public boolean removeEldestEntry(Map.Entry eldest) {
	            return size() > MAX_ENTRIES;
	        }
	    };
		
	}

	public void propertyChange(PropertyChangeEvent evt) {
		//UPDATE BUFFER DATA if it is for us
		if(OSCAddress.equals(evt.getPropertyName())) {
			synchronized(this) {
				recordedData.add((Float)evt.getNewValue());
				if(outputOsc) {
					Activator.getDefault().pcs.firePropertyChange(ExemplarConstants.PROP_NEW_FILTERED_DATA,Converter.a2dFloatToInt(getFilteredValueforTime(recordedData.size()-1)),oscOutputAddress);
				}
			}
		}
	}
	
	public Float getValueForTime(Integer t) {
		if(t<0 || t>=recordedData.size()) return 0.f; //TODO: exception
		else return recordedData.get(t);
	}
	
	public Float getFilteredValueforTime(Integer t) {
		// Note : Filter stack up so that higher index in the vector corresponding to higher placed filter in the stack.
		// Example: f(g(h(x))) -> index(f) < index(g) < index(h)
		if(t<0 || t>=recordedData.size()) {
			//we're accessing out of bounds
			//System.err.println("out of bounds access :"+t+" for "+recordedData.size());
			//return a value but don't write it to cache
			return 0.f;
		}
		//try to get the value from the cache first
		Float f = cache.get(t);
		if(f ==null) {
			//if we got a cache miss: calculate
			f = (float)Math.max(0.0, Math.min(1.0, this.recurseStackedFilters(t, this.filterList.size()-1)));
			//and add to cache
			cache.put(t,f);
			miss++;
		} else {
			hit++;
		}
		// print cache hit statistics
		if(++count%10000==0) {
			//System.err.println("Cache performance for sensor "+getOSCAddress()+": "+((float)hit/(float)count)+" (hit rate)");
			hit=0;miss=0;count=0;
			
		}
		return f;		
	}
	/**
	 * Return a value at an intermediate level of the filter stack
	 * @param time
	 * @param level
	 * @return
	 */
	public Float getFilteredValueforTime(int time, int level) {
		return this.recurseStackedFilters(time,level);
	}
	
	private Float recurseStackedFilters(Integer t, Integer currentTopLevel) {		
		/// TODO : Implement a non Non Recursive Filter Implementation. Will improve perf
		if(currentTopLevel < 0) {
		//	 base case for recursion
			return this.getValueForTime(t);
		}
		//if current filter is not active, pass through to next lowest level
		if(false == filterList.get(currentTopLevel).active) {
			return recurseStackedFilters(t, currentTopLevel-1);
		}
		//else: do some computation
		
		if(FilterType.DIFF.equals(filterList.get(currentTopLevel).type)) { 
			return 0.5f+(float)(recurseStackedFilters(t, currentTopLevel -1) - recurseStackedFilters(t-1, currentTopLevel -1));
		}
		else if (FilterType.SMOOTHE.equals(filterList.get(currentTopLevel).type)) {
			//TODO: this is SLOW!
			SmootheFilter smoothe;
			smoothe = new SmootheFilter(width, sd);		// Alternatively, statically set the Width and sd. Will not support 2 filters then
			Float result = 0.f;
			for (int i = 0 ; i < width; i++) {
				result += recurseStackedFilters(t-i,currentTopLevel-1) * smoothe.gauss(i);
			}
			return (result/smoothe.getArea());			
		} else if (FilterType.SCALE.equals(filterList.get(currentTopLevel).type)) {
			if(scaleFactor > 0) {
				if(scaleBaseline==ScaleBaseline.MIDDLE){
					return 0.5f + (recurseStackedFilters(t, currentTopLevel-1)-0.5f)*scaleFactor;
				}else {
					return recurseStackedFilters(t, currentTopLevel-1)*scaleFactor; 
				}
				
			} else { 
				if(scaleBaseline==ScaleBaseline.MIDDLE) {
					return 1.0f-(float)Math.abs(Math.min(0.0, Math.max(-1.0, (0.5f + (recurseStackedFilters(t, currentTopLevel-1)-0.5f)*scaleFactor))));
				} else {
					return 1.0f-(float)Math.abs(Math.min(0.0, Math.max(-1.0,recurseStackedFilters(t, currentTopLevel-1)*scaleFactor)));
				}
			}
		} else if (FilterType.OFFSET.equals(filterList.get(currentTopLevel).type)) {
			return recurseStackedFilters(t, currentTopLevel-1)+offset;
		} else if (FilterType.USERDEFINED.equals(filterList.get(currentTopLevel).type)) {
			FilterExtension f = FilterExtensionLoader.getInstance().getFilter(this,filterList.get(currentTopLevel).detail);
			if(f!=null) {
				f.setSensor(this);
				f.setLevel(currentTopLevel);
				return f.process(t);
			} else {
				System.err.println("Sensor.recurseStackedFilters badness");
				return 0.f;	
			}
			
		}
		/*case ExemplarConstants.MAXMINFILTER:
			// t must be atleast width time old
			// Vector<Float> diff;
			// diff = new Vector<Float>();
			Integer sameSignL = new Integer(0);
			Integer diffSignL = new Integer(0);
			Integer sameSignR = new Integer(0);
			Integer diffSignR = new Integer(0);
			Float sameSignDirectionL = new Float(0.0);
			Float sameSignDirectionR = new Float(0.0);
			float Temp = new Float(0.0);
			
			for (int i = 0; i < MAXMINWIDTH; i++) {
				if ((Temp = 
					 (recurseStackedFilters(t-i, currentTopLevel -1) 
						- recurseStackedFilters(t-i-1, currentTopLevel -1)))
						* (recurseStackedFilters(t-i-1, currentTopLevel -1) 
							- recurseStackedFilters(t-i-2, currentTopLevel -1))> 0) {
					sameSignL++;
					sameSignDirectionL = Temp;
				}
				else {
					diffSignL++;
				}
				if ((Temp = 
					 (recurseStackedFilters(t+i, currentTopLevel -1) 
						- recurseStackedFilters(t+i-1, currentTopLevel -1)))
						* (recurseStackedFilters(t+i+1, currentTopLevel -1) 
							- recurseStackedFilters(t+i, currentTopLevel -1))> 0) {
					sameSignR++;
					sameSignDirectionR = Temp;
				}
				else
					diffSignR++;
			}
			
			if (sameSignDirectionR * sameSignDirectionL < 0 && sameSignR > DIRECTIONTHRESH && sameSignL > DIRECTIONTHRESH)
			{
				return (Float) 1.f;
			}
			else
				return 0.f;
			*/
		else {
			return getValueForTime(t);
		}
	}
	
	public int getBufferSize() {
		return recordedData.size();
	}

	//TODO: these should be cached as well
	public Float getMinValueForTimeSlice(int tFrom, int tTo) {
		return getMinTimeValueForTimeSlice(tFrom, tTo).value;
	}

	public TimeValuePair getMinTimeValueForTimeSlice(int tFrom, int tTo) {
		float min= 1.f;
		int time=tFrom;
		for(int t=Math.max(0,tFrom); t<=Math.min(getBufferSize(),tTo); t++) {
			if (this.getFilteredValueforTime(t)<min) {
				min=this.getFilteredValueforTime(t);
				time=t;
			}
		}
		return new TimeValuePair(time,min);
	}
	
	/**
	 * 
	 * @param time
	 * @param width
	 * @return
	 */
	public TimeValuePair getMaxTimeValueForTimeSlice(int tFrom, int tTo) {
		float max= 0.f;
		int time=tFrom;
		for(int t=Math.max(0,tFrom); t<=Math.min(getBufferSize(),tTo); t++) {
			if (this.getFilteredValueforTime(t)>max) {
				max=this.getFilteredValueforTime(t);
				time=t;
			}
		}
		return new TimeValuePair(time,max);
	}
	
	public Float getMaxValueForTimeSlice(int tFrom, int tTo) {
		return getMaxTimeValueForTimeSlice(tFrom, tTo).value;
	}
			
	
	public String getOSCAddress() {
		return OSCAddress;
	}

		

	public void setScaleFactor(Float scaleFactor) {
		if(this.scaleFactor!=scaleFactor) {
			this.scaleFactor = scaleFactor;
			cache.clear();
		}
	}

	public void setOffset(float offset) {
		if(this.offset!=offset) {
			this.offset = offset;
			cache.clear();
		}
		
	}

	public void setSmoothe(int width, float sd) {
		if(this.width!=width || this.sd!= sd) {
			this.width=width;
			this.sd = sd;
			cache.clear();
		}
		
	}

	public void setActiveFilterVector(Vector<FilterEntry> v) {
		filterList = v;
		cache.clear();
	}

	public Vector<FilterEntry> getActiveFilterVector() {
		return filterList;
	}

	public float getOffset() {
		return offset;
	}

	public float getScaleFactor() {
		return scaleFactor;
	}

	public int getSmoothe() {
		return width;
	}
	
	/**
	 * Write object out to file
	 * @param out
	 * @throws IOException
	 */
	private void writeObject(ObjectOutputStream out) throws IOException {
		synchronized(this) {
			//uncomment to skip writing out all collected data
			//ArrayList<Float> temp = recordedData;
			//recordedData = new ArrayList<Float>();
			out.defaultWriteObject();
			//recordedData = temp;
		}
	}
	
	/**
	 * Read object back in from file and reregister with the propertychangesupport.
	 * 	@param in
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		Activator.getDefault().pcs.addPropertyChangeListener(OSCAddress,this);
		initCache();
	}

	public void clear() {
		cache.clear();
		recordedData.clear();
	}
	/**
	 * OSC output address
	 * @param text
	 */
	public void setOscOutputAddress(String text) {
	
		this.oscOutputAddress = text;
		outputOsc = (text!=null);

		
		
	}

	public String getOscOutputAddress() {
		return oscOutputAddress; 
	}

	public void setScaleBaseline(ScaleBaseline sb) {
		scaleBaseline=sb;
	}
	public ScaleBaseline getScaleBaseline() {
		return scaleBaseline;
	}

	public Object getRecordedData() {
		return recordedData;
	}

	public String getPseudocode() {
		// TODO generate string pseudocode based on filter stack
		// NOTE all we care about is the final filtered data here - 
		// we don't need to keep the source around.
		// can do either the recursion or something like
		// return offset(scale(rate(smooth(t))));
		String s = "getFilteredValue() {\n\treturn ";
		int numparens=0;
		for(FilterEntry f:filterList) { //TODO reverse direction; this is wrong
			if(f.active) {
				numparens++;
				if(f.type.equals(FilterType.DIFF)) {
					s+="diff(";
				}
				else if (f.type.equals(FilterType.OFFSET)) {
					s+="offset("+Math.round(offset*new Float(Activator.getDefault().getPreferenceStore().getInt(PreferenceConstants.P_MAXA2DVAL)))+",";
				}
				else if (f.type.equals(FilterType.SCALE)) {
					s+="scale("+Math.round(scaleFactor*1000)+",";
					if(scaleBaseline.equals(ScaleBaseline.BOTTOM))
						s+="true,";
					else
						s+="false,";
					
				}
				else if (f.type.equals(FilterType.SMOOTHE)) {
					s+="smoohte(";
				}
				else {
					s+="userfilter(";
				}
			}
		}
		s+="getValue()";
		for(int i=1;i<=numparens;i++) {
			s+=")";
		}
		s+=";\n}\n\n";
		return "//sensor.getPseudocode() not yet fully implemented.\n"+s+getFilterCode();
	}
	
		private String getFilterCode() {
		String s="";
		s+="int offset(int off, int val) {\n";
		s+="\treturn val+off;\n";
		s+="}\n\n";	
		
		s+="int scale(int scale, bool frombottom, int val) {\n";
		s+="\treturn ...;\n";
		s+="}\n\n";
		
		s+="int diff(int val) {\n";
		s+="\tstatic int last=0;\n";
		s+="\treturn 0.5f+val-last;\n";
		s+="\tlast=val;\n";
		s+="}\n\n";
		
		s+="int smoothe(int val) {\n";
		s+="\tstatic int last=0;\n";
		s+="\t//stuff to do here...\n";
		s+="}\n\n";
		return s;

	}
		
	public class TimeValuePair {
		int time;
		float value;
		public TimeValuePair(int time, float value) {
			super();
			this.time = time;
			this.value = value;
		}
		
	}
	/**
	 * Integrate sensor readings over a signal region (using math.abs)
	 * @param startTime
	 * @param endTime
	 * @return
	 */
	public float getPowerForTimeSlice(int startTime, int endTime) {
		float pwr=0;
		for(int t=startTime; t<endTime; t++) {
			pwr+=Math.abs(getFilteredValueforTime(t+1)-getFilteredValueforTime(t));
		}
		return pwr;
	}
}
