package edu.stanford.hci.exemplar;

import java.util.ArrayList;
import java.util.List;

import edu.stanford.hci.exemplar.preferences.PreferenceConstants;
/**
 * Exemplar
 * Copyright (c) 2004-2007, Stanford University
 * All rights reserved.
 *
 * <p>
 * This software is distributed under the <a href="http://hci.stanford.edu/research/copyright.txt">
 * BSD License</a>.
 * </p>
 * 
 * 
 * @author <a href="http://bjoern.org/">Bjoern Hartmann</a> ( bjoern(AT)stanford.edu )
 * @author Leith Abdulla
 */
public class ThresholdEventEvaluator implements EventEvaluator {
	
	private static final long serialVersionUID = 1L;
	private float min;
	private float max;
	private float minHysteresis;
	private float maxHysteresis;
	private int timeout; //how many sensor readings a signal has to be above/below (debounce) before event is reco'd
	private EventRegion growingRegion = null;
	private int updateInterval, updateWidthInterval;
	private Event event;
	private String oscAddress;
	private int lastStopTime=-1;
	private boolean haveFiredGrowingRegion = false;
	private EventRegion lastGrowingRegion=null;
	private enum TypeHint {NONE,EDGE,SLICE};
	private TypeHint typeHint = TypeHint.NONE;
	public ThresholdEventEvaluator(Event evt, String oscAddress) {
		super();
		event = evt;
		updateInterval=2;
		updateWidthInterval=2;
		max = 0.f;
		min = 1.f;
		timeout=1; //default timeout; was 10
		minHysteresis=0.f;
		maxHysteresis=0.f;
		this.oscAddress = oscAddress;
		
	}

	
	public boolean evaluate(int curTime, int width, boolean fireEvents) {
		if(max<=0.f && min>=1.f)
			return true;
		//System.err.println("ThresholdEvent.evaluate(): ["+(curTime-(width-1))+" - "+ curTime+"], "+oscAddress);
		//find if there is an event region at the last time before our window 
		//(we may want to extend it)
		if(growingRegion==null)
			growingRegion = findPrecedingRegion(curTime-width);
		
		Sensor s = Activator.getDefault().getSensorForOscAddress(oscAddress);
		float v;
		for(int t=curTime-(width-1); t<=curTime; t++) {
			v = s.getFilteredValueforTime(t);
			//threshold comparison with hysteresis:
			//this used to be: if(v>=min && v<=max) {
			if((growingRegion==null && v>=(min+minHysteresis) && v<=(max-maxHysteresis)) || //out going in
					( growingRegion!=null && v>=(min-minHysteresis) && v<=(max+maxHysteresis)) ) {//in staying in

				//within thresholds
				lastStopTime = -1;
				//start or grow a new region
				if(growingRegion==null) {
					//start a new region
					growingRegion = new EventRegion(t,t,oscAddress);
					growingRegion.setOrigin(EventRegionOrigin.SYSTEM_CREATED);
					haveFiredGrowingRegion =false;
				} else {
					growingRegion.setEndTime(t); // move end of existing region out by one
				}
				//if the region is as wide as our timeout,add it and fire event.
				if(!haveFiredGrowingRegion && growingRegion.getEndTime()-growingRegion.getStartTime() >= timeout) {
					haveFiredGrowingRegion=true;
					event.add(growingRegion);
					if(fireEvents) {
						event.evaluatorNotify(this,true);
						//Activator.getDefault().pcs.firePropertyChange(ExemplarConstants.PROP_EVENT_ON, null, event.key);
					}
				}
			} else {
				//not within thresholds
				//if in going out:
				if(growingRegion!=null) {
					// keep moving end time out for now
					growingRegion.setEndTime(t); // move end of existing region out by one
					if(lastStopTime<0) {
						//we just got below the threshold
						//save the timestamp
						lastStopTime = growingRegion.getEndTime();
						//reset growing region so it does not improperly fire/draw
						//lastGrowingRegion = growingRegion;
						//growingRegion=null;
					} else {
						//we've been out of threshold for a while. 
						//but long enough to stop event?
						if((t-lastStopTime) >= timeout) {
							if(fireEvents && haveFiredGrowingRegion) {
								event.evaluatorNotify(this,false);
								//Activator.getDefault().pcs.firePropertyChange(ExemplarConstants.PROP_EVENT_OFF, null, event.key);
							}

							// yes! so reset region
							growingRegion.setEndTime(lastStopTime);
							growingRegion=null;
							haveFiredGrowingRegion =false;
							// reset last stop time
							lastStopTime=-1;
							// and fire OFF event
						}
					}
				}
			}
		}
		return true;
	}

	private EventRegion findPrecedingRegion(int time) {
		//TODO: write a better than linear search thingy
		for(EventRegion r: event.getRegions()) {
			if(r.isPointInRegion(time) && oscAddress.equals(r.getAddress())) 
				return r;
		}
		return null;
	}

	public int getUpdateInterval() {
		return updateInterval;
	}

	public int getUpdateWidthInterval() {
		return updateWidthInterval;
	}

	public void learn(List<EventRegion> examples) {
		float min=1.f,max=0.f;
		float temp;
		// go through all regions
		for(EventRegion r:examples) {
			Sensor s = Activator.getDefault().getSensorForOscAddress(oscAddress);
			// check if regions max and min value are greater/smaller than current running max min.
			// if so, update
			
			temp = s.getMaxValueForTimeSlice(r.getStartTime(), r.getEndTime());
			if(temp>max)
				max =temp;
			temp = s.getMinValueForTimeSlice(r.getStartTime(), r.getEndTime());
			if(temp<min) {
				min =temp;
			}
		}
		this.max = max;
		this.min = min;
	}


	public float getMax() {
		return max;
	}


	public float getMin() {
		return min;
	}
	
	public void clear() {
		updateInterval=2;
		updateWidthInterval=2;
		max = 0.f;
		min = 1.f;
	}


	public EventType getType() {
		return EventType.THRESHOLD;
	}


	public void setMin(float f) {
		min = Math.max(0.f,Math.min(1.f,f));
		
	}

	public void setMax(float f) {
		max = Math.max(0.f,Math.min(1.f,f));
		
	}
	
	public void checkAndSwap() {
		if(min>max) {
			float temp = min;
			min=max;
			max=temp;
		}
	}


	public float getMinHysteresis() {
		return minHysteresis;
	}


	public float getMaxHysteresis() {
		return maxHysteresis;
	}


	public void setMaxHysteresis(float maxHysteresis) {
		this.maxHysteresis = maxHysteresis;
	}


	public void setMinHysteresis(float minHysteresis) {
		this.minHysteresis = minHysteresis;
	}


	public int getTimeout() {
		
		return timeout;
	}


	public void setTimeout(int timeout) {
		this.timeout = timeout;
	}

	// first attempt at writing pseudocode to export evaluator outside exemplar
	// haven't added hysteresis or timeouts yet
	public String getPseudocode() {
		String s = "";
		s+="bool evaluate() {\n";
		s+="\t//note: hysteresis and timeouts not yet supported.\n";
		s+="\tint v = getFilteredSensorValue();\n";
		s+="\tif(v>="+(min*new Float(Activator.getDefault().getPreferenceStore().getInt(PreferenceConstants.P_MAXA2DVAL)))+" && v<="+(max*new Float(Activator.getDefault().getPreferenceStore().getInt(PreferenceConstants.P_MAXA2DVAL)))+")\n";
		s+="\t\treturn true;\n";
		s+="\telse\n";
		s+="\t\treturn false;\n";
		s+="}\n";
		return s;
	}

	public boolean adjustRegion(EventRegion clutchRegion) {
		Sensor sensor = Activator.getDefault().getSensorForOscAddress(oscAddress);
		if(sensor==null) {
			System.err.println("null pointer in adjust region: oscaddr="+oscAddress+"; sensor="+sensor);
		}
		//r = region
		int rStart = clutchRegion.startTime;
		int rEnd = clutchRegion.endTime;
		//s = search
		int preSearchSamples = Math.round((1000.f/Activator.getDefault().getSamplingPeriodMs())*0.5f); //1/2 second
		int sStart = Math.max(0,rStart-preSearchSamples);
		
		
		Float sMin = Math.max(0.0f,sensor.getMinValueForTimeSlice(sStart, rStart-1)-.01f);
		Float sMax = Math.min(1.0f, sensor.getMaxValueForTimeSlice(sStart, rStart-1)+.01f);
		Float sAvg = (sMin+sMax)/2.f;
		
		Sensor.TimeValuePair minPair = sensor.getMinTimeValueForTimeSlice(rStart, rEnd);
		Sensor.TimeValuePair maxPair = sensor.getMaxTimeValueForTimeSlice(rStart, rEnd);
		Float rMin = Math.max(0.0f,minPair.value);
		Float rMax = Math.min(1.0f,maxPair.value);
		
		float dMin = Math.abs(sAvg-rMin);
		float dMax = Math.abs(rMax-sAvg);
		
		int newStartTime=rStart;
		int newEndTime=rEnd;
		int t;
		if(dMin>dMax) {
			// go for min
			// grow towards beginning
			for(t=minPair.time;t>=rStart;t-- ) {
				if(sensor.getFilteredValueforTime(t)>(sMin-(Math.abs(sMin-rMin)/2.f))) {
					newStartTime = t+1;
					break;
				}
			}
			// grow towards end
			for(t=minPair.time;t<=rEnd; t++) {
				if(sensor.getFilteredValueforTime(t)>(sMin-(Math.abs(sMin-rMin)/2.f))) {
					newEndTime = t-1;
					break;
				}
			}
		} else {
			//go for a max
			float v;
			for(t=maxPair.time;t>=rStart;t-- ) {
				v = sensor.getFilteredValueforTime(t);
				if(v<(sMax+(Math.abs(rMax-sMax)/2.f))) {
					newStartTime = t+1;
					break;
				}
			}
			// grow towards end
			for(t=maxPair.time;t<=rEnd; t++) {
				v = sensor.getFilteredValueforTime(t);
				if(v<(sMax+(Math.abs(rMax-sMax)/2.f))) {
					newEndTime = t-1;
					break;
				}
			}
		}
		clutchRegion.setStartTime(newStartTime);
		clutchRegion.setEndTime(newEndTime);
		return true;
	}
	
	/**
	 * Adjust region start and end parameters
	 * * find rmin,rmax of region before clutchRegion
	 * * find longest continuous segments lt, gt where val<rmin<rmax and rmin<rmax<val, respectively
	 * * if either lt or gt exceed a minimum length (50% of original clutchregion length),
	 *   set clutch region params to the longer one. otherwise, do nothing.
	 * @param clutchRegion
	 */
	public boolean adjustRegion2(EventRegion clutchRegion) {
		Sensor sensor = Activator.getDefault().getSensorForOscAddress(oscAddress);
		// sStart----rStart=====rEndsEnd
		//r = region
		int rStart = clutchRegion.startTime;
		int rEnd = clutchRegion.endTime;
		//s = search
		int preSearchSamples = Math.round((1000.f/Activator.getDefault().getSamplingPeriodMs())*0.5f); //1/2 second
		int sStart = Math.max(0,rStart-preSearchSamples);
		
		//Float sMin = Math.max(0.0f, sensor.getMinValueForTimeSlice(sStart, Math.min(sStart,rStart-1))-0.05f);
		//Float sMax = Math.min(1.0f, sensor.getMaxValueForTimeSlice(sStart, Math.min(sStart,rStart-1))+0.05f);
		
		Float sMin = Math.max(0.0f,sensor.getMinValueForTimeSlice(sStart, rStart-1)-.01f);
		Float sMax = Math.min(1.0f, sensor.getMaxValueForTimeSlice(sStart, rStart-1)+.01f);
		
		Float val;
		List<EventRegion> gtRegions = new ArrayList<EventRegion>();
		List<EventRegion> ltRegions = new ArrayList<EventRegion>();
		EventRegion gtRegion = null;
		EventRegion ltRegion = null;
		EventRegion maxGtRegion = null;
		EventRegion maxLtRegion = null;
		
		int maxGtSize=0;
		int maxLtSize=0;
		
		//ITERATE OVER MARKED REGION, CREATING LIST OF SUBREGIONS WITH VALUES THAT ARE LARGER AND SMALLER THAN
		//THE MIN,MAX OF THE SAMPLES PRECEDING THE MARKED REGION
		for(int t=rStart; t<=rEnd; t++) {
			val = sensor.getFilteredValueforTime(t);
			if(val>sMax) {
				ltRegion=null;
				if(gtRegion ==null) {
					gtRegion = new EventRegion(t,t,oscAddress);
					gtRegions.add(gtRegion);
				} else {
					gtRegion.setEndTime(t);
					if(gtRegion.getSize()>maxGtSize) {
						maxGtRegion = gtRegion;
						maxGtSize = gtRegion.getSize();
					}
						
				}
			} else if (val<sMin) {
				gtRegion=null;
				if(ltRegion ==null) {
					ltRegion = new EventRegion(t,t,oscAddress);
					ltRegions.add(ltRegion);
				} else {
					ltRegion.setEndTime(t);
					if(ltRegion.getSize()>maxLtSize) {
						maxLtRegion = ltRegion;
						maxLtSize = ltRegion.getSize();
					}
				}
			} else {
				gtRegion=null;
				ltRegion=null;
			}
		} //end for loop
		
		float minAdjustFrac = 0.01f; //adjusted region has to be at least 1% of original region size
		if(Math.max(maxLtSize,maxGtSize)>(clutchRegion.getSize()*minAdjustFrac )) {
			//adjustment successful, so change clutch region params
			if(maxLtSize>=maxGtSize) {
				clutchRegion.setStartTime(maxLtRegion.getStartTime());
				clutchRegion.setEndTime(maxLtRegion.getEndTime());
			} else {
				clutchRegion.setStartTime(maxGtRegion.getStartTime());
				clutchRegion.setEndTime(maxGtRegion.getEndTime());
			}
			return true;
		} else {
			//adjustment failed, do nothing;
			return false;
		}
	}
}
